App.Facilities.StatsHelper = class {
	/** Make a statistics table with given column labels
	 * @param {string[]} columns - Array of four labels for data columns. The first is wider than the others (typically used for facility revenue).
	 * @returns {HTMLTableElement}
	 */
	makeStatsTable(columns) {
		const table = document.createElement("table");
		table.border = "1";
		table.className = "facility-stats";
		const header = App.UI.DOM.appendNewElement("tr", table, null, ["header"]);
		App.UI.DOM.appendNewElement("th", header, "Items"); // first column, flexible width
		App.UI.DOM.appendNewElement("th", header, columns[0], ["wide"]); // "revenue" column, wider
		for (let i = 1; i < columns.length; ++i) {
			App.UI.DOM.appendNewElement("th", header, columns[i], ["narrow"]); // additional narrow columns
		}
		return table;
	}

	/** Adds a value row to a stats table, with the given label and value cells
	 * @param {HTMLTableElement} table
	 * @param {string|Node} label
	 * @param {HTMLTableCellElement[]} valueCells
	 * @param {boolean} [totals]
	 */
	addValueRow(table, label, valueCells, totals) {
		const row = App.UI.DOM.appendNewElement("tr", table, undefined, totals ? "total" : undefined);
		const labelCell = App.UI.DOM.appendNewElement("td", row, label);
		if (totals) {
			labelCell.style.fontWeight = "bold";
		}
		for (const cell of valueCells) {
			row.appendChild(cell);
		}
	}

	/** Makes a slave label
	 * @param {string} slaveName
	 * @param {string} customLabel
	 * @returns {Node}
	 */
	makeSlaveLabel(slaveName, customLabel) {
		if (!customLabel) {
			return document.createTextNode(slaveName);
		} else {
			const frag = document.createDocumentFragment();
			const label = App.UI.DOM.appendNewElement("span", frag, `(${customLabel}) `, "custom-label");
			frag.appendChild(label);
			frag.appendChild(document.createTextNode(slaveName));
			return frag;
		}
	}

	/** Makes a value cell
	 * @param {"reputation"|"cash"|"food"} type
	 * @param {number} value - numeric value
	 * @param {object} [flags]
	 * @param {boolean} [flags.forceNeg] - treat nonzero positive values as negative
	 * @param {boolean} [flags.showSign] - display the sign
	 * @param {boolean} [flags.showZero] - show the value, even if it's zero
	 * @param {boolean} [flags.bold] - style in bold
	 * @returns {HTMLTableCellElement}
	 */
	makeValueCell(type, value, flags = {}) {
		const cell = document.createElement("td");
		cell.classList.add("value");

		if (value !== 0 || flags.showZero) {
			// style appropriately
			if (value < 0 || flags.forceNeg) {
				cell.classList.add(type); // reputation or cash
				cell.classList.add("dec"); // loss
			} else if (value > 0) {
				cell.classList.add(type); // reputation or cash
				cell.classList.add("inc"); // gain
			}
			if (flags.bold) {
				cell.style.fontWeight = "bold";
			}

			// set contents
			let prefix = '';
			let suffix = '';
			if (type === "cash") {
				prefix += '¤';
			}
			if (type === "food") {
				suffix += 'kg';
			}
			if (flags.showSign) {
				if (flags.forceNeg) { // if the real value is negative, - sign will come from value.toFixed
					prefix += '-';
				} else if (value > 0) {
					prefix += '+';
				} else if (value === 0) {
					prefix += '±';
				}
			}
			let fixedPrecision = 1;
			if (type === "cash") {
				fixedPrecision = 2;
			} else if (type === "food") {
				fixedPrecision = 0;
			}
			const parts = value.toFixed(fixedPrecision).split('.');
			cell.appendChild(document.createTextNode(prefix + parts[0] + suffix));
			if (fixedPrecision > 0) {
				App.UI.DOM.appendNewElement("span", cell, '.' + parts[1], /^0+$/.test(parts[1]) ? ["decimalZero"] : undefined);
			} else {
				App.UI.DOM.appendNewElement("span", cell);
			}
		}

		return cell;
	}

	/** Make an empty cell (for parity)
	 * @returns {HTMLTableCellElement}
	 */
	makeEmptyCell() {
		return document.createElement("td");
	}

	/** Make the customer cell (for the facility-specific slave details column)
	 * @param {number} value
	 * @returns {HTMLTableCellElement}
	 */
	makeCustomersCell(value) {
		const cell = document.createElement("td");
		cell.classList.add("value");
		if (value <= 0) {
			cell.textContent = "none";
			cell.classList.add("red");
		} else {
			cell.textContent = value.toString();
		}
		return cell;
	}

	/** Make the production cell (for the facility-specific slave details column)
	 * @param {number} milk
	 * @param {number} cum
	 * @param {number} fluid
	 * @returns {HTMLTableCellElement}
	 */
	makeProductionCell(milk, cum, fluid) {
		const cell = document.createElement("td");
		cell.classList.add("value");
		if (milk || cum || fluid) {
			cell.textContent = `${milk}/${cum}/${fluid}`;
		} else {
			cell.textContent = `0/0/0`;
		}
		return cell;
	}

	/** Appends the slave stats row to the main stats table, and returns a reference to the slave stats table for later use.
	 * @param {HTMLTableElement} statsTable
	 * @param {string} caption
	 * @param {string[]} columns - six column labels - slave type, facility-specific output details column, wide column, then three narrow columns
	 * @returns {HTMLTableElement}
	 */
	makeSlaveStatsTable(statsTable, caption, columns) {
		const row = App.UI.DOM.appendNewElement("tr", statsTable);
		const bigCell = document.createElement("td");
		bigCell.colSpan = columns.length;
		const label = App.UI.DOM.appendNewElement("span", bigCell, caption);
		label.style.fontWeight = "bold";
		row.appendChild(bigCell);

		const table = document.createElement("table");
		table.className = "facility-stats-slave";
		const header = App.UI.DOM.appendNewElement("tr", table, "", "header");
		App.UI.DOM.appendNewElement("th", header, columns[0]); // first column, flexible width
		App.UI.DOM.appendNewElement("th", header, columns[1], "narrow"); // facility-specific output details
		App.UI.DOM.appendNewElement("th", header, columns[2], "wide"); // "revenue" column, wider
		for (let i = 3; i < columns.length; ++i) {
			App.UI.DOM.appendNewElement("th", header, columns[i], "narrow"); // additional narrow columns
		}
		bigCell.appendChild(table);
		statsTable.appendChild(bigCell);
		return table;
	}
};

App.Facilities.Brothel.Stats = (function() {
	const H = new App.Facilities.StatsHelper();
	const assureList = [ "whoreIncome", "rep", "whoreCosts", "adsIncome", "maintenance", "adsCosts", "totalIncome", "totalExpenses", "profit" ];

	/** Make the brothel slave revenue cell
	 * @param {number} revenue
	 * @param {number} adsIncome
	 * @returns {HTMLTableCellElement}
	 */
	function makeRevenueCell(revenue, adsIncome) {
		function printCash(value, cell) {
			const parts = value.toFixed(2).split('.');
			App.UI.DOM.appendNewElement("span", cell, '¤' + parts[0], ["cash", "inc"]);
			App.UI.DOM.appendNewElement("span", cell, '.' + parts[1], ["cash", "inc", "decimalPart"]);
		}
		const cell = document.createElement("td");
		cell.classList.add("value");
		printCash(revenue, cell);
		if (adsIncome > 0) {
			cell.appendChild(document.createTextNode(" ("));
			printCash(adsIncome, cell);
			cell.appendChild(document.createTextNode(" due to advertising)"));
		}
		return cell;
	}

	/** Generate the brothel statistics table
	 * @param {boolean} showDetails
	 * @returns {HTMLElement|DocumentFragment}
	 */
	function output(showDetails) {
		if (!V.showEconomicDetails) {
			return document.createDocumentFragment();
		} else if (!V.facility || !V.facility.brothel) {
			return App.UI.DOM.makeElement("h4", `- No statistics for ${V.brothelName} gathered this week -`);
		}

		const b = V.facility.brothel;
		for (const prop of assureList) {
			b[prop] = b[prop] || 0;
		}

		const stats = H.makeStatsTable(["Revenue", "Expenses", "Net Income", "Rep. Change"]);
		H.addValueRow(stats, "Total whoring income", [
			H.makeValueCell("cash", b.whoreIncome),
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreIncome),
			H.makeValueCell("reputation", b.rep, {showSign: true})
		]);
		H.addValueRow(stats, "Total whore living costs", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true}),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		if (showDetails) {
			const slaveStats = H.makeSlaveStatsTable(stats, "Whore details", ["Whore", "Customers", "Revenue", "Expenses", "Net Income", "Rep. Change"]);
			for (const record of b.income.values()) {
				const revenue = record.income + record.adsIncome;
				const netIncome = revenue - record.cost;
				H.addValueRow(slaveStats, H.makeSlaveLabel(record.slaveName, record.customLabel), [
					H.makeCustomersCell(record.customers),
					makeRevenueCell(revenue, record.adsIncome),
					H.makeValueCell("cash", record.cost, {forceNeg: true}),
					H.makeValueCell("cash", netIncome),
					H.makeValueCell("reputation", record.rep, {showSign: true})
				]);
			}
		}
		if (b.adsIncome > 0) {
			H.addValueRow(stats, "Additional income", [
				H.makeValueCell("cash", b.adsIncome),
				H.makeEmptyCell(),
				H.makeValueCell("cash", b.adsIncome),
				H.makeEmptyCell()
			]);
		}
		H.addValueRow(stats, "Brothel maintenance", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true}),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		if (b.adsCosts > 0) {
			H.addValueRow(stats, "Advertising program", [
				H.makeEmptyCell(),
				H.makeValueCell("cash", b.adsCosts, {forceNeg: true}),
				H.makeValueCell("cash", b.adsCosts, {forceNeg: true, showSign: true}),
				H.makeEmptyCell()
			]);
		}
		H.addValueRow(stats, "Total", [
			H.makeValueCell("cash", b.totalIncome),
			H.makeValueCell("cash", b.totalExpenses, {forceNeg: true}),
			H.makeValueCell("cash", b.profit, {bold: true}),
			H.makeValueCell("reputation", b.rep, {showSign: true, bold: true})
		], true);

		return stats;
	}

	return output;
})();

App.Facilities.Club.Stats = (function() {
	const H = new App.Facilities.StatsHelper();
	const assureList = [ "whoreIncome", "rep", "whoreCosts", "adsIncome", "maintenance", "adsCosts", "totalIncome", "totalExpenses", "profit" ];

	/** Make the slut efficiency cell
	 * @param {number} efficiency
	 * @param {boolean} [bold]
	 * @returns {HTMLTableCellElement}
	 */
	function makeEfficiencyCell(efficiency, bold) {
		const cell = document.createElement("td");
		cell.classList.add("value");
		if (bold) {
			cell.style.fontWeight = "bold";
		}
		const parts = efficiency.toFixed(2).split('.');
		const val = App.UI.DOM.appendNewElement("span", cell, parts[0], ["reputation", "inc"]);
		App.UI.DOM.appendNewElement("span", val, '.' + parts[1], /^0+$/.test(parts[1]) ? "decimalZero" : undefined);
		cell.appendChild(document.createTextNode(" rep/¤"));
		return cell;
	}

	/** Generate the Club statistics table
	 * @param {boolean} showDetails
	 * @returns {HTMLElement|DocumentFragment}
	 */
	function output(showDetails) {
		if (!V.showEconomicDetails) {
			return document.createDocumentFragment();
		} else if (!V.facility || !V.facility.club) {
			return App.UI.DOM.makeElement("h4", `- No statistics for ${V.clubName} gathered this week -`);
		}

		const b = V.facility.club;
		for (const prop of assureList) {
			b[prop] = b[prop] || 0;
		}

		const stats = H.makeStatsTable(["Rep. Gain", "Expenses", "Rep/Expenses", "Extra Income"]);
		H.addValueRow(stats, "Total slut rep gain", [
			H.makeValueCell("reputation", b.whoreIncome, {showSign: true}),
			H.makeEmptyCell(),
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.rep)
		]);
		H.addValueRow(stats, "Total slut living costs", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true}),
			H.makeEmptyCell(),
			H.makeEmptyCell()
		]);
		if (showDetails) {
			const slaveStats = H.makeSlaveStatsTable(stats, "Public slut details", ["Slut", "Customers", "Rep. Gain", "Expenses", "Rep/Expenses", "Extra Income"]);
			for (const record of b.income.values()) {
				const netIncome = record.income / record.cost;
				H.addValueRow(slaveStats, H.makeSlaveLabel(record.slaveName, record.customLabel), [
					H.makeCustomersCell(record.customers),
					H.makeValueCell("reputation", record.income, {showSign: true, showZero: true}),
					H.makeValueCell("cash", record.cost, {forceNeg: true}),
					makeEfficiencyCell(netIncome),
					H.makeValueCell("cash", record.rep, {showZero: true})
				]);
			}
		}
		if (b.adsIncome > 0) {
			H.addValueRow(stats, "Additional rep gain", [
				H.makeValueCell("reputation", b.adsIncome),
				H.makeEmptyCell(),
				H.makeEmptyCell(),
				H.makeEmptyCell()
			]);
		}
		H.addValueRow(stats, "Club maintenance", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true}),
			H.makeEmptyCell(),
			H.makeEmptyCell()
		]);
		if (b.adsCosts > 0) {
			H.addValueRow(stats, "Advertising program", [
				H.makeEmptyCell(),
				H.makeValueCell("cash", b.adsCosts, {forceNeg: true}),
				H.makeEmptyCell(),
				H.makeEmptyCell()
			]);
		}
		H.addValueRow(stats, "Total", [
			H.makeValueCell("reputation", b.totalIncome, {showSign: true}),
			H.makeValueCell("cash", b.totalExpenses, {forceNeg: true}),
			makeEfficiencyCell(b.profit, true),
			H.makeValueCell("cash", b.rep, {bold: true, showZero: true})
		], true);

		return stats;
	}

	return output;
})();

App.Facilities.Arcade.Stats = (function() {
	const H = new App.Facilities.StatsHelper();
	const assureList = [ "whoreIncome", "rep", "whoreCosts", "maintenance", "totalIncome", "totalExpenses", "profit" ];

	/** Generate the arcade statistics table
	 * @param {boolean} showDetails
	 * @returns {HTMLElement|DocumentFragment}
	 */
	function output(showDetails) {
		if (!V.showEconomicDetails) {
			return document.createDocumentFragment();
		} else if (!V.facility || !V.facility.arcade) {
			return App.UI.DOM.makeElement("h4", `- No statistics for ${V.arcadeName} gathered this week -`);
		}

		const b = V.facility.arcade;
		for (const prop of assureList) {
			b[prop] = b[prop] || 0;
		}

		const stats = H.makeStatsTable(["Revenue", "Expenses", "Net Income", "Rep. Change"]);
		H.addValueRow(stats, "Total arcade income", [
			H.makeValueCell("cash", b.whoreIncome),
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreIncome),
			H.makeValueCell("reputation", b.rep, {showSign: true})
		]);
		H.addValueRow(stats, "Total fuckmeat living costs", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true}),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		if (showDetails) {
			const slaveStats = H.makeSlaveStatsTable(stats, "Fuckmeat details", ["Fuckmeat", "Customers", "Revenue", "Expenses", "Net Income", "Rep. Change"]);
			for (const record of b.income.values()) {
				const revenue = record.income + record.adsIncome;
				const netIncome = revenue - record.cost;
				H.addValueRow(slaveStats, H.makeSlaveLabel(record.slaveName, record.customLabel), [
					H.makeCustomersCell(record.customers),
					H.makeValueCell("cash", revenue),
					H.makeValueCell("cash", record.cost, {forceNeg: true}),
					H.makeValueCell("cash", netIncome),
					H.makeValueCell("reputation", record.rep, {showSign: true})
				]);
			}
		}
		H.addValueRow(stats, "Arcade maintenance", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true}),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		H.addValueRow(stats, "Total", [
			H.makeValueCell("cash", b.totalIncome),
			H.makeValueCell("cash", b.totalExpenses, {forceNeg: true}),
			H.makeValueCell("cash", b.profit, {bold: true}),
			H.makeValueCell("reputation", b.rep, {showSign: true, bold: true})
		], true);

		return stats;
	}

	return output;
})();

App.Facilities.Dairy.Stats = (function() {
	const H = new App.Facilities.StatsHelper();
	const assureList = [ "whoreIncome", "whoreCosts", "maintenance", "totalIncome", "totalExpenses", "profit" ];

	/** Generate the arcade statistics table
	 * @param {boolean} showDetails
	 * @returns {HTMLElement|DocumentFragment}
	 */
	function output(showDetails) {
		if (!V.showEconomicDetails) {
			return document.createDocumentFragment();
		} else if (!V.facility || !V.facility.dairy) {
			return App.UI.DOM.makeElement("h4", `- No statistics for ${V.dairyName} gathered this week -`);
		}

		const b = V.facility.dairy;
		for (const prop of assureList) {
			b[prop] = b[prop] || 0;
		}

		const stats = H.makeStatsTable(["Revenue", "Expenses", "Net Income", "Rep. Change"]);
		H.addValueRow(stats, "Total cow income", [
			H.makeValueCell("cash", b.whoreIncome),
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreIncome),
			H.makeEmptyCell(),
		]);
		H.addValueRow(stats, "Total cow living costs", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true}),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		if (showDetails) {
			const slaveStats = H.makeSlaveStatsTable(stats, "Cow details", ["Cow", "Milk/Cum/Fluids", "Revenue", "Expenses", "Net Income", "Rep. Change"]);
			for (const record of b.income.values()) {
				const netIncome = record.income - record.cost;
				H.addValueRow(slaveStats, H.makeSlaveLabel(record.slaveName, record.customLabel), [
					H.makeProductionCell(record.milk, record.cum, record.fluid),
					H.makeValueCell("cash", record.income),
					H.makeValueCell("cash", record.cost, {forceNeg: true}),
					H.makeValueCell("cash", netIncome),
					H.makeEmptyCell()
				]);
			}
		}
		H.addValueRow(stats, "Dairy maintenance", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true}),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		H.addValueRow(stats, "Total", [
			H.makeValueCell("cash", b.totalIncome),
			H.makeValueCell("cash", b.totalExpenses, {forceNeg: true}),
			H.makeValueCell("cash", b.profit, {bold: true}),
			H.makeEmptyCell()
		], true);

		return stats;
	}

	return output;
})();

App.Facilities.Farmyard.Stats = (function() {
	const H = new App.Facilities.StatsHelper();
	const assureList = [ "whoreIncome", "whoreCosts", "maintenance", "totalIncome", "totalExpenses", "food", "profit" ];

	/** Generate the Farmyard statistics table
	 * @param {boolean} showDetails
	 * @returns {HTMLElement|DocumentFragment}
	 */
	function output(showDetails) {
		if (!V.showEconomicDetails) {
			return document.createDocumentFragment();
		} else if (!V.facility || !V.facility.farmyard) {
			return App.UI.DOM.makeElement("h4", `- No statistics for ${V.farmyardName} gathered this week -`);
		}

		const b = V.facility.farmyard;
		for (const prop of assureList) {
			b[prop] = b[prop] || 0;
		}

		const stats = H.makeStatsTable(["Revenue", "Expenses", "Food", "Net Income", "Rep. Change"]);
		H.addValueRow(stats, "Total farmhand income", [
			H.makeValueCell("cash", b.whoreIncome),
			H.makeEmptyCell(),
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreIncome),
			H.makeEmptyCell(),
		]);
		H.addValueRow(stats, "Total farmhand living costs", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true}),
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.whoreCosts, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		H.addValueRow(stats, "Total food produced", [
			H.makeEmptyCell(),
			H.makeEmptyCell(),
			H.makeValueCell("food", b.food),
			H.makeValueCell("food", b.food),
			H.makeEmptyCell()
		]);
		if (showDetails) {
			const slaveStats = H.makeSlaveStatsTable(stats, "Farmhand details", ["Farmhand", "Revenue", "Expenses", "Food", "Net Income", "Rep. Change"]);
			for (const record of b.income.values()) {
				const netIncome = record.income - record.cost;
				H.addValueRow(slaveStats, H.makeSlaveLabel(record.slaveName, record.customLabel), [
					H.makeValueCell("cash", record.income),
					H.makeValueCell("cash", record.cost, {forceNeg: true}),
					H.makeValueCell("food", record.food),
					H.makeValueCell("cash", netIncome),
					H.makeEmptyCell()
				]);
			}
		}
		H.addValueRow(stats, "Farmyard maintenance", [
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true}),
			H.makeEmptyCell(),
			H.makeValueCell("cash", b.maintenance, {forceNeg: true, showSign: true}),
			H.makeEmptyCell()
		]);
		H.addValueRow(stats, "Total", [
			H.makeValueCell("cash", b.totalIncome),
			H.makeValueCell("cash", b.totalExpenses, {forceNeg: true}),
			H.makeValueCell("food", b.food),
			H.makeValueCell("cash", b.profit, {bold: true}),
			H.makeEmptyCell()
		], true);

		return stats;
	}

	return output;
})();
